__author__ = 'wolkens'

import math, pygame

import const, myglobals, coloring, vector

class Point():
    def  __init__(self, x = 0.0, y = 0.0, z = 0.0): # Create a point object
        self.x, self.y, self.z  = x, y, z # Store coordinates
        self.px, self.py = 0, 0 # Projected to plane

    def setxyz(self, xyz): # Place point to new location
        self.x, self.y, self.z  = xyz[0], xyz[1], xyz[2] # Store coordinates

    def getxyz(self): # Read coordinates as tuple
        return (self.x, self.y, self.z)

    def projecttoplane(self, viewz): # By default, viewpoint is on the z axis
        zratio = -viewz / (self.z - viewz) # Ratio of projection
        self.px = math.trunc(zratio * self.x) # Calculate screen position
        self.py = math.trunc(zratio * self.y * const.VSCALE) # Right handed coordinate system, y axis points down if viewz < 0

class Vertex():
    def  __init__(self, size, xyz, pos, fixed):
        self.size = size # Scaling factor
        self.coords = Point(xyz[0]*size, xyz[1]*size, xyz[2]*size) # Set location
        self.offsetx, self.offsety, self.offsetz = pos, 0.0, const.OFFSETZ # Linear shift in space        
        self.fixed = fixed # Defines mode of operation
        self.velocity = 0.0 # Physical model variables
        self.attenuation = 0.70 # Shows how strong the vertex is attenuated
        self.accel = 0.0
        self.force = 0.25 # Shows how strong the vertex is accelerated
        self.lastlevel = 0.0 # Store last position for physical model
        self.springconst = 0.03 # As the name says
        self.springforce = 5.0 # No need to initialize actually :-)
        self.calibrate = 1.0
        self.drawsize = 0.0
        self.rotated = Point()
        self.rotated_sized = Point()
        self.overload = False
        self.delaycounter = 0

    def behavior(self, lvl, ol):
        if ol: # Overload event
            if self.delaycounter == 0: self.delaycounter = const.BLOWUPTIME + const.RELASETIME # Start overload
        if self.delaycounter > 0: # Overload active
            if self.delaycounter > const.RELASETIME: # Blowup phase
                self.drawsize = self.drawsize + (100 - self.drawsize) / 3 # Converge to 100
            else: # Release phase
                if self.fixed: self.drawsize = self.drawsize * 0.5 # Release fixed vertices faster
                else:
                    if self.delaycounter == const.RELASETIME:
                        self.drawsize = self.drawsize * 2.5 # Blow up at first
                    if self.delaycounter < const.RELASETIME - 3: # Delayed release for moving vertices
                        self.drawsize = self.drawsize * 0.8 # Release moving vertices slower
                self.velocity, self.level, self.lastlevel = 0.0, 0.0, 0.0 # Full reset
            self.delaycounter = self.delaycounter - 1
        else: # Not in overload state
            if not self.fixed: # Update moving vertices
                self.level = lvl
                self.springforce = self.springconst * self.lastlevel # Spring tension
                self.accel = self.force * (self.level - self.lastlevel) \
                    - self.springforce # Vertex acceleration
                self.velocity += self.accel # Velocity changes
                self.lastlevel += self.velocity # Needle is in motion
                self.velocity *= self.attenuation # Velocity is attenuated
                self.drawsize = self.lastlevel * self.calibrate # Position to draw at
        return # For debug only

    def rotate(self, rm): # rm for rotation matrix
        self.rotated.x = rm[0][0]*self.coords.x + rm[0][1]*self.coords.y + rm[0][2]*self.coords.z
        self.rotated.y = rm[1][0]*self.coords.x + rm[1][1]*self.coords.y + rm[1][2]*self.coords.z
        self.rotated.z = rm[2][0]*self.coords.x + rm[2][1]*self.coords.y + rm[2][2]*self.coords.z
        ds161 = 1 + self.drawsize/160
        self.rotated_sized.setxyz( (self.rotated.x * ds161 + self.offsetx, 
        self.rotated.y * ds161 + self.offsety, self.rotated.z * ds161 + self.offsetz) )
        self.rotated_sized.projecttoplane(const.VIEWZ) # Project to screen

    def getrotatedcoords(self):
        return (self.rotated_sized.x, self.rotated_sized.y, self.rotated_sized.z)

    def getprojectedcoords(self):
        return (self.rotated_sized.px, self.rotated_sized.py)

class Triangle():  # Triangle of 3 Vertices
    def __init__(self, Alabel, Blabel, Clabel, r = 128, g = 128, b = 128, vertices = {} ):  # Create a triangle from 3 vertices
        self.Alabel, self.Blabel, self.Clabel = Alabel, Blabel, Clabel # Store vertex label and coordinates
        self.red, self.green, self.blue, self.linevalue = r, g, b, 255 # Default line color is white
        self.vertexlist = vertices
        self.visible = True  # Assume visible
        self.drawred, self.drawgreen, self.drawblue, self.drawlinevalue = 128, 128, 128, 255  # Actual drawing colors
        self.lightray = Point(1 / math.sqrt(3), 1 / math.sqrt(3), -1/math.sqrt(3))  # The sun shines from left top
        self.Center = Point(0, 0, 0) # Center of triangle
        self.nx, self.ny, self.nz, self.nlength = 0.0, 0.0, 0.0, 0.0  # Normal vector of triangle
        self.sp_color = 0.0

    def update(self): # Update the triangle
	    # Calculate center point
        Ap = self.vertexlist[self.Alabel].rotated_sized
        Bp = self.vertexlist[self.Blabel].rotated_sized
        Cp = self.vertexlist[self.Clabel].rotated_sized
        self.Center.x = (Ap.x + Bp.x + Cp.x ) / 3 
        self.Center.y = (Ap.y + Bp.y + Cp.y ) / 3 
        self.Center.z = (Ap.z + Bp.z + Cp.z ) / 3 
        # Calculate normal vector immediately
        Vx, Vy, Vz = vector.vector_subtract(self.vertexlist[self.Blabel].getrotatedcoords(), 
		    self.vertexlist[self.Alabel].getrotatedcoords() ) # The AB vector
        Wx, Wy, Wz = vector.vector_subtract(self.vertexlist[self.Clabel].getrotatedcoords(), 
		    self.vertexlist[self.Alabel].getrotatedcoords() ) # The AC vector
        self.nx = Vy * Wz - Vz * Wy # Calculate normal vector components
        self.ny = Vz * Wx - Vx * Wz
        self.nz = Vx * Wy - Vy * Wx
        self.nlength = vector.vector_length( (self.nx, self.ny, self.nz) )
        self.nx, self.ny, self.nz = vector.vector_scale( (self.nx, self.ny, self.nz), 1/self.nlength) # Normalize vector
        # Calculate actual color if seen from viewpoint
        # Vector pointing from triangle middle to viewpoint
        vx, vy, vz = vector.vector_subtract( (0, 0, const.VIEWZ), self.Center.getxyz() )
        vlength = vector.vector_length( (vx, vy, vz) )  # Calculate vector length
        vx, vy, vz = vector.vector_scale( (vx, vy, vz) , 1/vlength ) # Normalize viewing vector
        sp_visible = vector.vector_scalarproduct( (vx, vy, vz), (self.nx, self.ny, self.nz) ) # Angle of view
        if sp_visible < 0: self.visible = False  # Triangle pointing away
        else:
            self.visible = True  # Triangle facing to viewer
            ux, uy, uz = vector.vector_subtract( (vx, vy, vz), self.lightray.getxyz() )
            # u bisects the angle between the light source and the viewing vector
            ulength = vector.vector_length( (ux, uy, uz) ) # Calculate vector length
            ux, uy, uz = vector.vector_scale( (ux, uy, uz) , 1/ulength )# Normalize bisecting vector
            self.sp_color = vector.vector_scalarproduct( (ux, uy, uz) , (self.nx, self.ny, self.nz) ) # Angle of light
            self.drawred, self.drawgreen, self.drawblue,  self.drawlinevalue = \
                myglobals.colorizer.coloringfunction(
                self.sp_color * 0.999, self.red, self.green, self.blue, self.linevalue) # 0 <= sp < 1

    def draw(self, dest, xoffs, yoffs, isred):  # Draw the triangle with outer border
        if isred:
            self.drawred, self.drawgreen, self.drawblue, self.drawlinevalue = myglobals.colorizer.coloringfunction(
                self.sp_color * 0.999, 255, 0, 0, self.linevalue) # 0 <= sp < 1
        Ap = self.vertexlist[self.Alabel].rotated_sized
        Bp = self.vertexlist[self.Blabel].rotated_sized
        Cp = self.vertexlist[self.Clabel].rotated_sized
        pygame.draw.polygon(dest, pygame.Color(coloring.rgb_to_hex(self.drawred, self.drawgreen, self.drawblue)),
			( (xoffs + Ap.px, yoffs + Ap.py), (xoffs + Bp.px, yoffs + Bp.py), (xoffs + Cp.px, yoffs + Cp.py) ) )
        pygame.draw.polygon(dest, pygame.Color(coloring.rgb_to_hex(self.drawlinevalue, self.drawlinevalue, self.drawlinevalue)),
			( (xoffs + Ap.px, yoffs + Ap.py), (xoffs + Bp.px, yoffs + Bp.py), (xoffs + Cp.px, yoffs + Cp.py) ), 3 )
